"""entry point for the paragon CLI."""
# pylint: disable = no-value-for-parameter

from typing import List
import click
from paragon import Paragon
from paragon.exporter import Exporter
from paragon.mark import Mark
from paragon.utils import Utils


@click.command()
@click.argument("code", nargs=-1)
@click.option(
    "--accuracy",
    "-a",
    required=False,
    default=1,
    help="Number of iterations",
)
@click.option(
    "--files",
    "-f",
    required=False,
    multiple=True,
    help="Input file path(s)",
    type=click.Path(),
)
@click.option(
    "--output",
    "-o",
    required=False,
    nargs=1,
    help="Output file path",
    type=click.Path(),
)
def cli(code: List[str], accuracy: int, files: List[str], output: str):
    """entry point for the paragon CLI."""

    # nothing to benchmark
    if len(code) == 0 and len(files) == 0:
        click.echo("You must provide python code to benchmark.", err=True)
        return

    start, current_benchmark = Mark(), [1]

    if code:
        bench_code(code, current_benchmark, accuracy)

    if files:
        bench_files(files, current_benchmark, accuracy)

    if output:
        Exporter.export_results(output)

    print("✨ Done in {:.2f} s ✨".format(start.diff()))


def bench_code(code: List[str], current_benchmark: List[int], accuracy: int):
    """runs benches for all passed in code"""
    for val in code:
        try:
            Paragon.bench(data=(val, val, *current_benchmark), accuracy=accuracy)
        except (NameError, SyntaxError) as error:
            click.echo(f"Error: {error}", err=True)

        current_benchmark[0] += 1


def bench_files(files: List[str], current_benchmark: List[int], accuracy: int):
    """runs benches for all passed in files"""
    for file in files:
        res, name, status = Utils.verify_file(file)

        if not status:
            click.echo(res + "\n", err=True)
            continue

        try:
            Paragon.bench(data=(res, name, *current_benchmark), accuracy=accuracy)
        except (NameError, SyntaxError) as error:
            click.echo(f"Error: {error}", err=True)

        current_benchmark[0] += 1


if __name__ == "__main__":
    cli()



"""all benchmark related logic"""
# pylint: disable = exec-used

import sys
import click
from paragon.stats import Stats
from paragon.utils import Utils
from paragon.mark import Mark
from paragon.animate import Animate


class Paragon:
    """handles code benchmarking"""

    results = []

    @staticmethod
    def benchmark(name: str = None, accuracy: int = 1):
        """benchmark decorator
        - retrieves decoratored functions name and args
        - calls bench with the function call and scope

        :param name: name of benchmark
        :accuracy: number of iterations
        """

        def inner(func):
            def wrap(*args):
                nonlocal name

                start, func_name = Mark(), func.__name__

                try:
                    click.secho(
                        "Benchmark: {}".format(name or func_name),
                        fg="white",
                        bold=True,
                    )
                except AttributeError:
                    click.secho(
                        "Make sure you wrap your recursive functions!\n", err=True
                    )
                    sys.exit(1)

                # gather locals and build code string
                env = {func_name: func}
                code = f"{func_name}(*{args})"

                # run bench
                Paragon.bench(data=(code, func_name, None), env=env, accuracy=accuracy)

                print(f"✨ Done in {round(start.diff(), 2)} s ✨\n")

                return func(*args)

            return wrap

        return inner

    @staticmethod
    def bench(data: tuple, env: object = None, accuracy: int = 1):
        """benchmark function
        - benches passed in code, taking in optional locals and number
        of iterations

        :param data: ( code: str, name: str, iteration: int )
        :accuracy: number of iterations
        :env: locals
        """
        code, name, iteration = data

        if iteration:
            click.secho(
                f"Benchmark #{iteration}: {name}",
                fg="white",
                bold=True,
            )

        # don't need stdout
        Utils.redirect_stdout()

        # run once to make sure it's valid python
        err, status = Utils.run_once(code, env)
        if not status:
            raise err

        mark, animate, times = Mark(), Animate(accuracy), []
        for _ in range(accuracy):
            exec(code, globals(), env or globals())
            animate.next()
            times.append(mark.diff())
            mark.reset()

        # cleanup animation and output stats
        animate.done()
        Paragon.results.append(Stats(times, name).output())



"""all mark related tests"""

import time
from paragon.mark import Mark


def test_subtract():
    """test mark subtraction"""
    start = Mark()
    time.sleep(2)
    end = Mark()
    assert round(end - start) == 2


def test_diff():
    """test mark diff"""
    start = Mark()
    time.sleep(2)
    assert round(start.diff()) == 2


def test_reset():
    """test mark reset"""
    start = Mark()
    time.sleep(3)
    start.reset()
    time.sleep(3)
    assert round(start.diff()) == 3



"""all utility related logic"""
# pylint: disable = exec-used

import os
import sys


class Utils:
    """utility methods"""

    @staticmethod
    def reset_stdout():
        """set stdout back to normal"""
        sys.stdout = sys.__stdout__

    @staticmethod
    def redirect_stdout(redirect_value=None):
        """set stdout to provided value or None"""
        sys.stdout = redirect_value

    @staticmethod
    def get_filename_ext(path: str):
        """returns the name and extension of the last part in the path"""
        parts, ext = os.path.splitext(path)
        return (parts.split("/")[-1], ext)

    @staticmethod
    def verify_file(path):
        """verifies a python file to benchmark"""
        if not os.path.exists(path):
            return ("File does not exist.", "", False)

        name, ext = Utils.get_filename_ext(path)
        if ext != ".py":
            return ("File must be a python file.", False)

        return (open(path, "r").read(), name, True)

    @staticmethod
    def run_once(code, env: object = None):
        """run code once to make sure it's valid"""
        try:
            exec(code, globals(), env or globals())
            return (None, True)
        except (NameError, SyntaxError) as error:
            Utils.reset_stdout()
            return (error, False)

    @staticmethod
    def format_to_int(fmt: str):
        """converts a file format to integer"""
        table = {".md": 1, ".json": 2, ".csv": 3}
        return table[fmt]



"""handles all exporting logic"""
import csv
import enum
import json
import click
from paragon import Paragon
from paragon.utils import Utils


class Format(enum.Enum):
    """file types as enums"""

    MARKDOWN = 1
    JSON = 2
    CSV = 3


class Exporter:
    """handles writing to various file formats"""

    @staticmethod
    def export_results(path: str):
        """writes results to a file"""
        _, ext = Utils.get_filename_ext(path)

        if ext not in [".md", ".json", ".csv"]:
            click.echo("Output file must be of type markdown, csv or json.\n", err=True)
            return

        with open(path, "w+") as out:
            Exporter.write(Format(Utils.format_to_int(ext)), out)

        click.echo(f"Successfully exported results to {path}\n")

    @staticmethod
    def write(fmt: Format, out: str):
        """builds and writes different outputs based on file format"""

        results = {
            result["name"]: {"average": result["average"], "range": result["range"]}
            for result in Paragon.results
        }

        if fmt == Format.JSON:
            json.dump(results, out, indent=4)

        if fmt == Format.MARKDOWN:
            table = "| Program | Average [s] | Min [s] | Max [s] |\n|---|---|---|---|\n"
            for key, val in results.items():
                table += f"| `{key}` | {val['average']} | {val['range'][0]} | {val['range'][1]} |\n"
            out.write(table)

        if fmt == Format.CSV:
            writer = csv.DictWriter(
                out, fieldnames=["Program", "Average [s]", "Min [s]", "Max [s]"]
            )

            writer.writeheader()

            for key, val in results.items():
                writer.writerow(
                    {
                        "Program": key,
                        "Average [s]": val["average"],
                        "Min [s]": val["range"][0],
                        "Max [s]": val["range"][1],
                    }
                )
